package Tabu;

import VoteReveal.VRPDataReader;
import VoteReveal.RevealOrder;
import VoteReveal.RevealProblem;

import java.util.*;

public class MultiObjectiveTabuSearch {
    // Algorithm variables
    private static RevealProblem problem;
    private static float[] mins;
    private static float[] eqs;
    private static int numNoUpdate;
    private static int solutionsSinceUpdate;

    // args: filename, funcs, size, tenure, neighbourhood, numNoUpdate
    public static void main(String[] args) {
        assert args.length > 5;
        String filename = args[0];
        char[] funcs = args[1].toCharArray();
        Arrays.sort(funcs); // Ensure consistent filenames when writing results
        int sz = Integer.parseInt(args[2]);
        int tenure = Integer.parseInt(args[3]);
        char neighbour = args[4].toCharArray()[0];
        numNoUpdate = Integer.parseInt(args[5]);
        
        problem = VRPDataReader.readData(filename);

        // create a set of random feasible solutions
        RevealOrder[] x = new RevealOrder[sz];
        for (int i = 0; i < x.length; i++) {
            x[i] = problem.getRandomSolution();
        }

        // Initialise tabu histories
        TabuList[] l = new TabuList[sz];
        for (int i = 0; i < l.length; i++) {
            l[i] = new TabuList(tenure);
        }

        // initialise empty set of non-dominated solutions
        Set<RevealOrder> ND = new HashSet<>();

        // calculate range equalisation factors for each objective
        rangeEqualisationFactors(funcs);

        int count = 0;
        solutionsSinceUpdate = 0;
        double startTime = System.currentTimeMillis();
        while (!stoppingCriterion()) {
            for (int i = 0; i < x.length; i++) {

                // Create weight vector (lambda) for Xi
                float[] weights = new float[funcs.length];
                boolean weightsAreAssigned = false;

                for (int j = 0; j < x.length; j++) {
                    if (i != j) {
                        // Get objective values for both solutions
                        float[] fXi = new float[funcs.length];
                        float[] fXj = new float[funcs.length];
                        for (int n = 0; n < funcs.length; n++) {
                            fXi[n] = x[i].getCost(funcs[n]);
                            fXj[n] = x[j].getCost(funcs[n]);
                        }

                        // Assign objective weights
                        if (!x[i].dominates(x[j], funcs)) {
                            float w = influence(distance(fXi, fXj, eqs));
                            for (int k = 0; k < funcs.length; k++) {
                                if (fXi[k] > fXj[k]) {
                                    weights[k] += w * eqs[k];
                                    weightsAreAssigned = true;
                                }
                            }
                        }
                    }
                }
                if (!weightsAreAssigned) weights = randomWeights(weights.length);
                normalise(weights);

                // Find the neighbout (Yi) of Xi which maximises weights*f(Yi) and which isn't tabu
                RevealOrder[] Neighbours = switch(neighbour) {
                    case 'R' -> x[i].getNeighbours(weights, funcs, mins, eqs);
                    case 'F' -> x[i].getFullNeighbours(weights, funcs, mins, eqs);
                    default -> throw new IllegalStateException("Unexpected neighbourhood: " + neighbour);
                };
                RevealOrder Yi = null;
                for (RevealOrder s : Neighbours) {
                    if (!l[i].contains(s.getAction())) {
                        Yi = s;
                        break;
                    }
                }
                assert Yi != null;

                // Add action to tabu list
                l[i].add(Yi.getAction());

                // Replace Xi with Yi
                x[i] = Yi;

                // Add Yi to the ND set only if it is not dominated by any of the solutions in the set
                boolean dominated = false;
                for (RevealOrder s : ND) {
                    if (s.dominates(Yi, funcs)) {
                        dominated = true;
                        break;
                    }
                }
                if (!dominated) {
                    // Remove any points in the ND set which are now dominated by Yi
                    solutionsSinceUpdate = 0;
                    List<RevealOrder> toBeRemoved = new ArrayList<>();
                    for (RevealOrder s : ND) {
                        if (Yi.dominates(s, funcs)) toBeRemoved.add(s);
                    }
                    // Remove after iteration to avoid ConcurrentModificationException
                    for (RevealOrder r : toBeRemoved) ND.remove(r);
                    ND.add(Yi);
                }
                else {
                    solutionsSinceUpdate++;
                }
            }
            count++;
        }
        double stopTime = System.currentTimeMillis();
        for (RevealOrder s : ND) {
            s.write(funcs, "MOTS_"+neighbour);
        }
        System.out.println(count);
        System.out.printf("Runtime: %fs\n", (stopTime-startTime)/1000);
    }

    // Scale an array of floats so that they sum to 1
    private static void normalise(float[] weights) {
        float sum = 0;
        for (float w : weights) sum += w;
        for (int i = 0; i < weights.length; i++) {
            weights[i] /= sum;
        }
    }

    // Generate a random weights array
    private static float[] randomWeights(int length) {
        Random r = new Random();
        float[] out = new float[length];
        for (int i = 0; i < out.length; i++)
            out[i] = r.nextFloat();
        return out;
    }

    // Nearby solutions should influence each other more than far solutions
    private static float influence(float d) {
        if (d == 0) return Integer.MAX_VALUE;
        else return 1/d;
    }

    // Get estimate of distance between the points given by two solutions
    // (manhatten distance norm)
    private static float distance(float[] fXi, float[] fXj, float[] ranges) {
        // Scale + sum differences between objective values for solutions
        float dist = 0;
        for (int i = 0; i < ranges.length; i++) {
            float diff = Math.abs(fXi[i] - fXj[i]);
            diff /= ranges[i];
            dist += diff;
        }
        return dist;
    }

    private static boolean stoppingCriterion() {
        return solutionsSinceUpdate >= numNoUpdate;
    }

    private static void rangeEqualisationFactors(char[] funcs) {
        float[] maxs = new float[funcs.length];
        mins = new float[funcs.length];
        eqs = new float[funcs.length];

        for (int i = 0; i < funcs.length; i++) {
            maxs[i] = Integer.MIN_VALUE;
            mins[i] = Integer.MAX_VALUE;
        }

        // Get ranges using 100,000 solutions
        for (int i = 0; i < 100000; i++) {
            RevealOrder x = problem.getRandomSolution();
            for (int j = 0; j < funcs.length; j++) {
                float val = x.getCost(funcs[j]);
                maxs[j] = Math.max(maxs[j], val);
                mins[j] = Math.min(mins[j], val);
            }
        }

        for (int i = 0; i < eqs.length; i++) {
            eqs[i] = 1/(maxs[i] - mins[i]);
        }
    }
}