package VoteReveal;

import java.util.Arrays;

public class EntertainmentEvaluation {
    public static float evaluate(int[] order, int[][] votes, int[][] cumulativeVals, char func) {
        int numCandidates = votes[0].length;
        return switch (func) {


            // Uncertainty
            case 'U' -> {
                int total = 0;
                for (int[] row : cumulativeVals) {
                    int[] rowCopy = new int[row.length];
                    System.arraycopy(row, 0, rowCopy, 0, rowCopy.length);
                    Arrays.sort(rowCopy);
                    total += Math.abs(rowCopy[1] - rowCopy[0]);
                    total += Math.abs(rowCopy[numCandidates - 1] - rowCopy[numCandidates - 2]);

                    for (int i = 1; i < rowCopy.length-1; i++) {
                        total += Math.min(
                                Math.abs(rowCopy[i] - rowCopy[i - 1]),
                                Math.abs(rowCopy[i] - rowCopy[i + 1])
                        );
                    }
                }
                yield -total; // Invert output so that function can be treated as maximising
            }


            // Biased uncertainty (only considers uncertainty between leaders)
            case 'B' -> {
                int total = 0;
                for (int[] row : cumulativeVals) {
                    int[] rowCopy = new int[row.length];
                    System.arraycopy(row, 0, rowCopy, 0, rowCopy.length);
                    Arrays.sort(rowCopy);
                    reverse(rowCopy);

                    total += Math.abs(rowCopy[0] - rowCopy[1]);
                    int goTo = Math.min(rowCopy.length-1, 5);
                    for (int i = 1; i < goTo; i++) {
                        total += Math.min(
                                Math.abs(rowCopy[i] - rowCopy[i - 1]),
                                Math.abs(rowCopy[i] - rowCopy[i + 1])
                        );
                    }
                }
                yield -total; // Invert output so that function can be treated as maximising
            }


            // Time-biased uncertainty (Gives greater weight to later stages of reveal)
            case 'T' -> {
                int total = 0;
                for (int r = 0; r < cumulativeVals.length; r++) {
                    int[] row = cumulativeVals[r];
                    int[] rowCopy = new int[row.length];
                    System.arraycopy(row, 0, rowCopy, 0, rowCopy.length);
                    Arrays.sort(rowCopy);
                    total += r * Math.abs(rowCopy[1] - rowCopy[0]);
                    total += r * Math.abs(rowCopy[numCandidates - 1] - rowCopy[numCandidates - 2]);

                    for (int i = 1; i < rowCopy.length-1; i++) {
                        total += r * Math.min(
                                Math.abs(rowCopy[i] - rowCopy[i - 1]),
                                Math.abs(rowCopy[i] - rowCopy[i + 1])
                        );
                    }
                }
                yield -total; // Invert output so that function can be treated as maximising
            }


            // Lead (between 1st and 2nd)
            case 'L' -> {
                int total = 0;
                float expected = 0;
                for (int i = 0; i < cumulativeVals.length; i++) {
                    // Get the max and second-max values from current row of candidate points
                    int max1 = Integer.MIN_VALUE;
                    int max2 = Integer.MIN_VALUE;
                    for (int j = 0; j < numCandidates; j++) {
                        if (cumulativeVals[i][j] > max1) {
                            max2 = max1;
                            max1 = cumulativeVals[i][j];
                        } else if (cumulativeVals[i][j] > max2) {
                            max2 = cumulativeVals[i][j];
                        }
                    }

                    // Get the expected points using the current voter
                    for (int j = 0; j < numCandidates; j++)
                        expected += votes[order[i]][j];
                    expected /= numCandidates;

                    // If the difference between max1 and max2 is less than the expected points
                    // per candidate there is no clear lead (increment total)
                    if (max1 - max2 <= expected)
                        total++;
                }
                yield total;
            }


            // Unique pole switching measure
            case 'P' -> {
                int[] flags = new int[numCandidates];
                for (int[] row : cumulativeVals) {
                    int maxVal = Integer.MIN_VALUE;
                    int maxIdx = 0;
                    for (int i = 0; i < row.length; i++) {
                        if (row[i] > maxVal) {
                            maxVal = row[i];
                            maxIdx = i;
                        }
                    }
                    flags[maxIdx] = 1; // Flag this candidate
                }
                int total = 0;
                for (int i : flags) total += i; // Count flagged candidates
                yield total;
            }


            // Pole switching measure
            case 'Q' -> {
                int total = 0;
                int prev = -1;
                for (int[] row : cumulativeVals) {
                    int maxVal = Integer.MIN_VALUE;
                    int maxIdx = 0;
                    for (int i = 0; i < row.length; i++) {
                        if (row[i] > maxVal) {
                            maxVal = row[i];
                            maxIdx = i;
                        }
                    }
                    // Check if pole has changed and increment counter
                    if (prev != -1 && maxIdx != prev) total++;
                    prev = maxIdx;
                }
                yield total;
            }

            default -> 0;
        };
    }

    private static void reverse(int[] array) {
        for (int i = 0; i < array.length / 2; i++) {
            int tmp = array[i];
            array[i] = array[array.length-i-1];
            array[array.length-i-1] = tmp;
        }
    }

    public static int[][] calculateCumulativeVals(int[] order, int[][] votes) {
        int numVoters = votes.length;
        int numCandidates = votes[0].length;
        int[][] out = new int[numVoters][numCandidates];

        // r = current reveal
        for (int r = 0; r < numVoters; r++) {
            // c = current candidate
            for (int c = 0; c < numCandidates; c++) {
                // for each candidate, initialise total as equal to previous total (if one exists)
                if (r > 0) out[r][c] = out[r-1][c];
                // add points from this reveal. order[r] = voter
                out[r][c] += votes[order[r]][c];
            }
        }
        return out;
    }
}
