package VoteReveal;

import Utils.Action;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

public class RevealOrder {
    private final int[] order;
    private final int[][] cumulativeVals;
    private final RevealProblem problem;
    private final Map<Character, Float> costs = new HashMap<>();
    private final Action action;

    // Initialise as a standalone solution
    public RevealOrder(int[] order, RevealProblem problem) {
        this.order = order.clone();
        this.problem = problem;
        cumulativeVals = EntertainmentEvaluation.calculateCumulativeVals(order, problem.votes);
        action = null;
    }

    // Initialise as a neighbour
    public RevealOrder(int[] oldOrder, Action a, RevealProblem problem) {
        this.problem = problem;
        action = a;
        this.order = oldOrder.clone();

        int tmp = order[a.i];
        order[a.i] = order[a.j];
        order[a.j] = tmp;

        cumulativeVals = EntertainmentEvaluation.calculateCumulativeVals(order, problem.votes);
    }

    // weak dominance: Xi dominates Xj when f(Xi) >= f(Xj) for all objectives
    // and f(Xi) > f(Xj) for at least one objective.
    public boolean dominates(RevealOrder x, char[] funcs) {
        boolean dominates = false;
        for (char func : funcs) {
            float fX = getCost(func);
            float fY = x.getCost(func);

            if (fX < fY) return false; // return early (first criteria failed)
            if (fX > fY) dominates = true;
        }
                                // this will be reached if all fXi >= fYi
        return dominates;       // second critera: is at least one fXi > fYi ?
    }

    public float getCost(char func) {
        if (costs.containsKey(func))
            return costs.get(func);
        else {
            float val = EntertainmentEvaluation.evaluate(order, problem.votes, cumulativeVals, func);
            costs.put(func, val); // Save in hashmap to avoid repeat calculations
            return val;
        }
    }

    // Takes a list of legal actions, generates neighbours,
    // and returns a sorted array of neighbouring solutions
    public RevealOrder[] getNeighbours(List<Action> actions, char func) {
        List<RevealOrder> neighbours = new ArrayList<>();
        for (Action a : actions) {
            neighbours.add(new RevealOrder(order, a, problem));
        }
        return neighbours.stream().sorted(new Comparator<RevealOrder>() {
            @Override
            public int compare(RevealOrder o1, RevealOrder o2) {
                float diff = o2.getCost(func) - o1.getCost(func);
                if (diff > 0) return 1;
                else if (diff < 0) return -1;
                else return 0;
            }
        }).toArray(RevealOrder[]::new);
    }

    // Return neighbour solutions sorted by their weighted cost of various functions
    public RevealOrder[] getNeighbours(float[] weights, char[] funcs, float[] mins, float[] eqs) {
        List<RevealOrder> neighbours = new ArrayList<>();
        for (int i = 0; i < order.length-1; i++) {
            neighbours.add(new RevealOrder(order, new Action(i, i+1), problem));
        }
        return neighbours.stream().sorted(new Comparator<RevealOrder>() {
            @Override
            public int compare(RevealOrder o1, RevealOrder o2) {
                float diff = o2.getWeightedCost(funcs, weights, mins, eqs) - o1.getWeightedCost(funcs, weights, mins, eqs);
                if (diff > 0) return 1;
                else if (diff < 0) return -1;
                else return 0;
            }
        }).toArray(RevealOrder[]::new);
    }

    public RevealOrder[] getFullNeighbours(float[] weights, char[] funcs, float[] mins, float[] eqs) {
        List<RevealOrder> neighbours = new ArrayList<>();
        for (int i = 0; i < order.length-1; i++) {
            for (int j = i+1; j < order.length; j++) {
                neighbours.add(new RevealOrder(order, new Action(i, j), problem));
            }
        }
        return neighbours.stream().sorted(new Comparator<RevealOrder>() {
            @Override
            public int compare(RevealOrder o1, RevealOrder o2) {
                float diff = o2.getWeightedCost(funcs, weights, mins, eqs) - o1.getWeightedCost(funcs, weights, mins, eqs);
                if (diff > 0) return 1;
                else if (diff < 0) return -1;
                else return 0;
            }
        }).toArray(RevealOrder[]::new);
    }

    private float getWeightedCost(char[] funcs, float[] weights, float[] mins, float[] eqs) {
        float sum = 0;
        for (int i = 0; i < funcs.length; i++) {
            float val = getCost(funcs[i]);
            val -= mins[i];
            val *= eqs[i];
            val *= weights[i];
            sum += val;
        }
        return sum;
    }

    public Action getAction() {
        return action;
    }

    public RevealOrder copy() {
        return new RevealOrder(order.clone(), problem);
    }

    @Override
    public String toString() {
        return Arrays.toString(order) +
                "\nU " +
                this.getCost('U') +
                "\nB " +
                this.getCost('B') +
                "\nT " +
                this.getCost('T') +
                "\nP " +
                this.getCost('P') +
                "\nQ " +
                this.getCost('Q') +
                "\nL " +
                this.getCost('L');
    }

    public void write(char func, String msg) {
        char[] funcs = new char[1];
        funcs[0] = func;
        write(funcs, msg);
    }

    public void write(char[] funcs, String msg) {
        System.out.println(costs.get(funcs[0]));
        Arrays.sort(funcs);
        String filename = problem.instanceFile;
        int end = filename.lastIndexOf(".");
        filename = filename.substring(0, end).concat("_");
        for (char c : funcs) filename = filename.concat(String.valueOf(c));
        filename = filename.concat(".csv");

        try {
            FileWriter fw = new FileWriter(filename, true);
            BufferedWriter bw = new BufferedWriter(fw);

            for (char func : funcs) {
                bw.write(Float.toString(getCost(func)));
                bw.write(",");
                bw.write(func);
                bw.write(",");
            }
            bw.write(msg + ",");
            String orderString = Arrays.toString(order);
            orderString = orderString.replace(" ", "");
            bw.write(orderString.substring(1, orderString.length()-1));

            bw.newLine();
            bw.close();
            fw.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public int[][] getCumulativeVals() {
        return cumulativeVals;
    }

    public String getCandidate(int i) {
        return problem.candidates[i];
    }
}
